#include <stdlib.h>
#include <sys/ipc.h>
#include <stdio.h>
#include <sys/shm.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>

			// SIMULATED CLOCK STRUCT // 
typedef struct {
	//seconds
        int clockS;
	//nanoseconds
        int clockN;
} shared_sum_t;
//create struct to go in shared memory
static shared_sum_t *sharedsum;


			// TIMED TERMINATION SIGNAL HANDLER //
int quitFlag = 0;
/* ARGSUSED */
static void myhandler(int s) {
 int errsave;
 errsave = errno;
 quitFlag = 1;
 printf("\n\nReached 2 Second Limit @ %d Seconds %d Nanoseconds in simulated time\nCleaning up and Terminating\n", sharedsum->clockS, sharedsum->clockN);
 errno = errsave;
}
static int setupinterrupt(void) { /* set up myhandler for SIGPROF */
 struct sigaction act;
 act.sa_handler = myhandler;
 act.sa_flags = 0;
 return (sigemptyset(&act.sa_mask) || sigaction(SIGPROF, &act, NULL));
}
static int setupitimer(void) { /* set ITIMER_PROF for 2-second intervals */
 struct itimerval value;
 value.it_interval.tv_sec = 2;
 value.it_interval.tv_usec = 0;
 value.it_value = value.it_interval;
 return (setitimer(ITIMER_PROF, &value, NULL));
}

			// HANDLER FOR CATCHING CTRL-C //
void sigintHandler(int sig_num) 
{ 
    quitFlag = 1;
    printf("\n\nCtrl-C was pressed @ %d Seconds %d Nanoseconds in simulated time\nCleaning up and Terminating\n", sharedsum->clockS, sharedsum->clockN);
    signal(SIGINT, sigintHandler); 
} 

int main(int argc, char *argv[]){

			// SET UP SIGNAL HANDLERS //
	
	if (setupinterrupt() == -1) {
 		perror("Failed to set up handler for SIGPROF");
 		return 1;
 	}
 	if (setupitimer() == -1) {
 		perror("Failed to set up the ITIMER_PROF interval timer");
 		return 1;
 	}
	//catches ctrl-c
	signal(SIGINT, sigintHandler); 
			// PARSE CL INPUT //
	//flags/variables to keep track of CL input
	int hflag = 0;
	int iflag = 0;
	int oflag = 0;
	int sflag = 0;
	int nflag = 0;
	int s;
	int n;
	int opt;	
	//handle input/output files
	FILE *inFile = NULL;
	FILE *outFile = NULL;
	const char *input;
	const char *output;

	//detect and store CL input
	while ((opt = getopt (argc, argv, "hi:o:s:n:")) != -1)
		switch (opt){
		  case 'h':
			hflag = 1;
		  	break;
                  case 'i':
                        iflag = 1;
			input = optarg;
			break;
                  case 'o':
                        oflag = 1;
			output = optarg;
			break;  
                  case 's':
                        sflag = 1;
                        s = atoi (optarg);
                        break;
                  case 'n':
                        nflag = 1;
                        n = atoi (optarg);
                        break;

		case '*':
			break;
		 default:
		  abort ();
		}
	
	//error check running environment
	if(s > n){
		printf("ERROR: invalid input. \n");
		printf("Total number of processes cannot be smaller than currently running processes. \n");
		return EXIT_SUCCESS;
	}
	
	
	//display help information and terminate
	if(hflag != 0){
		printf("\nThis program is intended to take an input file consisting of an amount\n");
                printf("to increment and times for child processes to launch and then terminate at.\n");
                printf("The input file is structured as:\n");
                printf("<amount to increment by>\n");
                printf("<second start time> <nanosecond start time> <duration before terminating>\n");
                printf("The program also takes in amount of processes to run <-n> and the amount\n");
                printf("that can run in the system at once <-s>. Input and output files are specified\n");
                printf("via <-i> and <-o> respectively. \n");
                printf("The input file must have enough lines of time specification to match the amount\n");
                printf("processes you intend to run <n>\n");
                printf("The default running environment: -i input.txt -o output.txt -n 4 -s 2\n");
                printf("For this to work you must have an input.txt already created.\n");
                printf("An example of successful program CL input would be: \n");
                printf("./oss -i inputFile -o outputFile -n 10 -s 5\n\n");
		return EXIT_SUCCESS;
	}
	//if both input & output files are specified
	if(iflag != 0 && oflag != 0){
		if ((inFile = fopen(input, "r")) == NULL){
		        perror(argv[0]);
		        printf("Error opening input file");
       			return EXIT_FAILURE;
		}
                if ((outFile = fopen(output, "w")) == NULL){
                        perror(argv[0]);
                        printf("Error opening output file");
                        return EXIT_FAILURE;
                }
	}
	//only input file specified
	else if(iflag != 0){
                if ((inFile = fopen(input, "r")) == NULL){
                        perror(argv[0]);
                        printf("Error opening input file");
                        return EXIT_FAILURE;
                }
                if ((outFile = fopen("output.txt", "w")) == NULL){
                        perror(argv[0]);
                        printf("Error opening output file");
                        return EXIT_FAILURE;
                }
	
	}
	//only outputfile specified
	else if(oflag != 0){
                if ((inFile = fopen("input.txt", "r")) == NULL){
                        perror(argv[0]);
                        printf("Error opening input file");
                        return EXIT_FAILURE;
                }
		if ((outFile = fopen(output, "w")) == NULL){
                        perror(argv[0]);
                        printf("Error opening output file");
                        return EXIT_FAILURE;
                }
	}
	//no input/output files specified
	else if(iflag == 0 && oflag == 0){
                if ((inFile = fopen("input.txt", "r")) == NULL){
                        perror(argv[0]);
                        printf("Error opening input file");
                        return EXIT_FAILURE;
                }
                if ((outFile = fopen("output.txt", "w")) == NULL){
                        perror(argv[0]);
                        printf("Error opening output file");
                        return EXIT_FAILURE;
                }

        }


	//set default running environment
	if(sflag == 0){	
		s = 2;
	}
	if(nflag == 0){
		n = 4;
	}

			// END OF CL PARSING //

	//read increment amount
        int increment;
        fscanf(inFile,"%d ",&increment);
	        


			// SET UP SHARED MEMORY //
	//generate key for shared memeory
	key_t key = ftok("oss.c", 50);
	if(key == -1){
		perror("key generation failed:");
		return EXIT_FAILURE;
	}

	//create shared memory segment and get id of it	
	int shmid = shmget(key,sizeof(shared_sum_t),0666|IPC_CREAT); 
	//check if failed creation
	if ((shmid == -1) && (errno != EEXIST))
		return -1;
	else{
		//attach to shared memory
		sharedsum = (shared_sum_t *) shmat(shmid,NULL,0);
		//check if failed
		if (sharedsum == (void *)-1){
			return -1;
		}
		//initialize simulated clock values to zero
		sharedsum -> clockS = 0;
		sharedsum -> clockN = 0;

	}
			// END OF SHARED MEM SET UP //
	
	//process parameters
	int running = 0;
	int count = 1;
	
	//process IDs
	pid_t childPid;
	pid_t test;
	int status;
	int pids[n];
	pids[0] = -255;
	int pcount = 0;
	//file input data
	int startS;
       	int startN;
        char endN[15];
	int rflag = 0;
	//ensure all children terminate
	int terminated = 0;
	int ntemp = n;
				// INCREMENT CLOCK AND LAUNCH PROCESSES //

	//run until desired processes have reached and all have terminated
	while (n != 0 || terminated != ntemp){
		//increment clock nanoseconds
                sharedsum -> clockN += increment;
		//increment clock seconds
                if(sharedsum -> clockN == 1000000000){
                	sharedsum -> clockN = 0;
			sharedsum -> clockS++;
                }
                if(running >= 20){
                        printf("\nReached 20 child process capacity\n");
                        quitFlag = 1;
                }

		//if pass the two second mark abandan and clean up
		if(quitFlag == 1){
			if(pids[0]=-255){
				break;
			}
			int i;
			for(i = 0; i < ntemp; i++){
				kill(pids[i], SIGKILL);
			}
			break;
		}

		//check if max running child processes has been reached
		if(running < s && n != 0){
			//read start and termination time once per process
			if(rflag == 0){
                        	fscanf(inFile,"%d %d %s\n", &startS, &startN, endN);
				rflag=1;
			}

			//check clock to see if child should launch
			if(sharedsum->clockN >= startN && sharedsum->clockS >= startS){
				//set read flag
				rflag = 0;
				//fork
				childPid = fork();
				//check for fork error
				if (childPid == -1){
					perror("failed to fork");
					return EXIT_FAILURE;
				}
				if (childPid == 0){
						// CHILD CODE //	
					//exec to user process
					execl("./user", endN, NULL);
					perror("exec failed");
					exit(0);
						// END OF CHILD CODE //
			
				}
				else{
        			                // PARENT CODE  //
        			                
        		               	fprintf(outFile, "Child Process(%d): %d Launched   @ %d Seconds %d Nanoseconds\n", count, childPid, sharedsum->clockS, sharedsum -> clockN);
	               			pids[pcount] = childPid;                 
					pcount++;
					//increment process count
					count++;

					//increment the amount of child processes currently running
					running++;
					n--;			
        		                	// END OF PARENT CODE //
				}//end of else

			}//end of if second stmt

		
		}//end of if process has ended

		//check if any children have terminated
		test = waitpid(-1, &status, WNOHANG);
                if(test > 0){
			terminated++;
			fprintf(outFile, "Child Process   : %d Terminated @ %d Seconds %d Nanoseconds\n", test, sharedsum->clockS, sharedsum->clockN);
                        running--;
                }
		
	} //end of while loop

	if(quitFlag == 0){
		printf("All processes launched and terminated successfully. \n");
	}

	fprintf(outFile,"\nSimulated clock reached: %d Seconds %d Nanoseconds\n", sharedsum->clockS, sharedsum->clockN);
        fclose(inFile);
        fclose(outFile);


	//detach from shared memory segment
	shmdt(sharedsum);

	//delete shared memory segment
	shmctl(shmid,IPC_RMID,NULL); 
	
	return EXIT_SUCCESS;
}
